from .error_handler import ErrorHandler
from .frontend import create_ratelimit_decorator, create_ratelimit_depends, create_ratelimit_middleware
from .limit_config import LimitConfig
from .limiter_engine import LimiterEngine


class Limiter:

    def __init__(
            self,
            engine: LimiterEngine,
            error_handler: ErrorHandler
    ):
        self.engine = engine
        self.error_handler = error_handler

    @property
    def decorator(self):
        return create_ratelimit_decorator(self.engine, self.error_handler)

    @property
    def depends(self):
        return create_ratelimit_depends(self.engine, self.error_handler)

    def middleware(self, config: LimitConfig):
        return create_ratelimit_middleware(self.engine, self.error_handler, config)



from typing import Optional, Union, Callable, Type

from sweelimiter.api.limiter_result import LimiterResult

ErrorHandlerFn = Callable[[LimiterResult], None]
ExceptionFactory = Callable[[LimiterResult], Exception]


def factory_to_handler_fn(factory: ExceptionFactory) -> ErrorHandlerFn:
    def handler_fn(result) -> None:
        raise factory(result)

    return handler_fn


class ErrorHandler:

    def __init__(self, handler_fn: ErrorHandlerFn):
        self._fn = handler_fn

    def handle(self, result: LimiterResult):
        self._fn(result)


class ErrorHandlers:

    @staticmethod
    def raise_exception(factory: ExceptionFactory):
        return ErrorHandler(factory_to_handler_fn(factory))

    @staticmethod
    def consume(fn: ErrorHandlerFn):
        return ErrorHandler(fn)



from typing import List, Union
from durations import Duration

from sweelimiter.api.interval import Interval

IntervalArgument = Union[str, Interval, int]


def get_interval(arg: IntervalArgument) -> Interval:
    if type(arg) == str:
        return Interval(arg)
    if type(arg) == int:
        return Interval.of_seconds(arg)
    if type(arg) == Interval:
        return arg
    raise Exception('Invalid interval:', arg)


class Limit:

    def __init__(
            self,
            interval: IntervalArgument,
            max_requests: int
    ):
        self.interval = get_interval(interval)
        self.max_requests = max_requests

    def __repr__(self):
        return f'Limit(interval={self.interval},max_requests={self.max_requests})'


class LimitConfig:

    def __init__(
            self,
            key: str,
            limits: List[Limit],
    ):
        self.key = key
        self.limits = limits

    def __repr__(self):
        return f'LimitConfig(key={self.key},limits={self.limits})'



from typing import Type

from fastapi import HTTPException
from requests import Response
from starlette.middleware.base import BaseHTTPMiddleware, RequestResponseEndpoint
from starlette.requests import Request

from sweelimiter.api.error_handler import ErrorHandler
from sweelimiter.api.limit_config import LimitConfig
from sweelimiter.api.limiter_engine import LimiterEngine


def create_ratelimit_middleware(
        engine: LimiterEngine,
        error_handler: ErrorHandler,
        config: LimitConfig
) -> Type[BaseHTTPMiddleware]:
    class LimiterMiddleware(BaseHTTPMiddleware):
        async def dispatch(self, request: Request, call_next: RequestResponseEndpoint) -> Response:

            limit_result = await engine.limit(config, request)

            if limit_result.limit_applied:
                if error_handler:
                    error_handler.handle(limit_result)
                else:
                    raise HTTPException(status_code=429, detail=limit_result.description)

            return await call_next(request)

    return LimiterMiddleware



from durations import Duration


class Interval(Duration):

    def __init__(self, repr: str, *args, **kwargs):
        super().__init__(repr, *args, **kwargs)

    @staticmethod
    def of_seconds(seconds: int):
        return Interval(str(seconds) + 's')

    @staticmethod
    def of_minutes(minutes: int):
        return Interval.of_seconds(minutes * 60)

    @staticmethod
    def of_hours(hours: int):
        return Interval.of_seconds(hours * 60 * 60)

    @staticmethod
    def of_days(days: int):
        return Interval.of_seconds(days * 60 * 60 * 24)



from typing import List, Union

from starlette.requests import Request

from .backend import Backend
from .bypass_checker import BypassChecker
from .key_extractor import KeyExtractor
from .limit_config import LimitConfig
from .limiter_result import LimiterResult


class LimiterEngine:

    def __init__(
            self,
            extractor: KeyExtractor,
            backend: Backend,
            bypass_checkers: List[BypassChecker] = None,
            prefix: str = 'global',
    ):
        self._prefix = prefix
        self._extractor = extractor
        self._backend = backend

        if not bypass_checkers:
            bypass_checkers = []

        self._bypass_checkers = bypass_checkers

    def _is_bypassed(self, req: Request) -> bool:
        if any(checker(req) for checker in self._bypass_checkers):
            return True
        else:
            return False

    async def limit(self, config: LimitConfig, req: Request) -> LimiterResult:
        if self._is_bypassed(req):
            return LimiterResult(False, 'Bypassed', [])

        limiter_meta = []
        error_hit: Union[LimiterResult, None] = None

        for req_key in self._extractor.extract(req):

            for limit in config.limits:
                full_key = f'{self._prefix}@{config.key}@{req_key}@{int(limit.interval.to_miliseconds())}'

                error, meta = await self._backend.process(full_key, limit.interval, limit.max_requests)

                limiter_meta.append(meta)

                if error is not None:
                    error_hit = LimiterResult(True, error, limiter_meta)

        if error_hit:
            return error_hit

        return LimiterResult(False, 'Limiters not hit', limiter_meta)



from .limit_config import Limit
from .limiter import Limiter
from .limiter_engine import LimiterEngine
from .error_handler import ErrorHandlers
from .interval import Interval
from .limiter_result import LimiterResult


from typing import List

from starlette.requests import Request

from .host_extractor import host_key_extractor
from ..api.key_extractor import KeyExtractor, KeyExtractorFn


def multi_extractor_fn_factory(extractors: List[KeyExtractor]) -> KeyExtractorFn:
    def fn(req: Request) -> List[str]:
        result = []
        for extractor in extractors:
            keys = extractor.extract(req)
            result.extend(keys)
        return result

    return fn


class Extractors:
    @staticmethod
    def multiple(extractors: List[KeyExtractor]):
        return KeyExtractor('', multi_extractor_fn_factory(extractors))

    @staticmethod
    def new(prefix: str, fn: KeyExtractorFn):
        return KeyExtractor(prefix + '@', fn)

    @staticmethod
    def host():
        return host_key_extractor



from functools import wraps
from typing import List

from fastapi import HTTPException
from starlette.requests import Request

from sweelimiter.api.error_handler import ErrorHandlerFn, ErrorHandler
from sweelimiter.api.limiter_engine import LimiterEngine
from sweelimiter.api.limit_config import LimitConfig, Limit


def create_ratelimit_decorator(engine: LimiterEngine, error_handler: ErrorHandler):
    def ratelimit(key: str, *args: Limit):
        config = LimitConfig(key, args)

        def decorator(func):
            @wraps(func)
            async def wrapper(request: Request, *args, **kwargs):
                limit_result = await engine.limit(config, request)

                if limit_result.limit_applied:
                    if error_handler:
                        error_handler.handle(limit_result)
                    else:
                        raise HTTPException(status_code=429, detail=limit_result.description)

                await func(request, *args, **kwargs)

            return wrapper

        return decorator

    return ratelimit
